#include <fstream>
#include <iostream>
#include <sstream> 
#include "matching.hh"
#include "lorentz.hh"

namespace TemplateOverlap{
  


/// Make sure that template subcones do not overlap.
bool checkTemplate (const jet_t & probe_template)
{  
  bool isOK = true; 
  for (jet_t::const_iterator i = probe_template.begin() ; i != probe_template.end() && isOK ; ++i)  {  
    for (jet_t::const_iterator j = i + 1 ; j != probe_template.end(); ++j){	
      if((*i).deltaR2((*j)) < 4 * (*i).radius * (*i).radius)
	{ isOK = false; break ;}
    }
  } 
  return (isOK);
}

/// Reset jet flags.
void reset(jet_t & jet){
  for (jet_t::iterator j = jet.begin(); j != jet.end(); ++j) 
    if ((*j).isUsed) (*j).isUsed = false;	
        
}


/// Sum pT around template momenta and return overlap.                  
double overlapDistance (jet_t & jet1, jet_t & jet2, double R0) {
double sum = 0.;
    for(jet_t::const_iterator k = jet1.begin(); k != jet1.end(); ++k) {
            double eTjetNow = 0.;              
       //  Sum up unused particles within required distance of parton.
    for (jet_t::iterator j = jet2.begin(); j !=jet2.end(); ++j) {
      if ((*j).isUsed) continue;
    if ((*j).deltaR2(*k) > R0 * R0 ) continue;
         eTjetNow += (*j).pT;
        (*j).isUsed = true;
    }              
    double eTcenterNow = (*k).pT;
    //change sigma to E/3
    sum += 4.5 * (eTjetNow - eTcenterNow) * (eTjetNow - eTcenterNow)/(eTcenterNow * eTcenterNow);

}

reset(jet2);
return (exp(-sum));
    }


void matchTemplate(jet_t & jet, const  vector<jet_t> & templates, vector<double>  & result, int match_method, int normalization)
{ 
  for (int i = 0; i < int (templates.size()); ++i){
      double sum = 0.;

    for(jet_t::const_iterator t_particle = templates[i].begin(); t_particle != 
      templates[i].end(); ++t_particle) 
    {
      double eTjetNow = 0.;
      //  Sum up unused particles within required distance of parton.
      for ( jet_t::iterator particle = jet.begin(); particle != jet.end(); ++particle) 
      {
	if ((*particle).isUsed) continue;
	if ((*particle).deltaR2(*t_particle) > (*t_particle).radius * (*t_particle).radius ) continue;
        eTjetNow += (*particle).pT;
	(*particle).isUsed = true;
      }
      double eTcenterNow = (*t_particle).pT;

    // change sigma to E/3
      sum += 4.5 * (eTjetNow - eTcenterNow) * (eTjetNow - eTcenterNow)/(eTcenterNow * eTcenterNow); 
    // if one overlap is too low, we skip the rest. 
      if (sum > 4.6 ) break;
      
    } 
// Clean flags. 
  reset(jet);
    
  result.push_back(exp(-sum));
  }
  

}

void maximize(const vector<double>  & result, double & maxVal, int & maxLoc )
{
   maxVal = 0;
   maxLoc = 0;

 for (int i = 0; i < int (result.size()); ++i){
                double thisVal = result[i];
		if (thisVal > maxVal){ maxLoc = i; maxVal = thisVal;}
 }
}

// Small number to avoid division by zero.
const double Event::TINY = 1e-20;

//--------------------------------------------------------------------------


 
/// Compute planar flow jet shape.
double Event::planarFlow() const{
	
  double Ixy = 0.,Ixx = 0.,Iyy = 0.;
  for ( int i = 0; i < size(); ++i) {
  //Define a orthogonal basis.
    Vec4 v1 = cross3(axis(), Vec4(0,0,1,0)); //Should check axis != z?
    Vec4 v2 = cross3(axis(), v1);
    v1 *= 1/sqrt(dot3(v1, v1));
    v2 *= 1/sqrt(dot3(v2, v2));
  //Get x and y components in that basis.
    double px=dot3(particles[i],v1),py=dot3(particles[i],v2),ee=particles[i].e();
  // Calculate matrix to be used in the computation of planar flow
    Ixx += px * px/ee;
    Iyy += py * py/ee;
    Ixy += px * py/ee;				
  }
//Return planar flow.
return(4 * (Ixx * Iyy -Ixy * Ixy)/((Ixx + Iyy )*(Ixx + Iyy )) );
}

/// Compute planar flow jet shape. Boost invariant way. 
double Event::planarFlow2(double r) const{
  double Ixy = 0.,Ixx = 0.,Iyy = 0.;
  double etaJet = axis().eta();
  double phiJet = axis().phi();

  for ( int i = 0; i < size(); ++i) {
    //Get x and y components in that basis.
    double  etaNow = particles[i].eta() - etaJet , phiNow = particles[i].phi() - phiJet,
            ee = particles[i].pT();
    if (abs(phiNow) > M_PI) phiNow = 2 * M_PI - abs(phiNow);
    // Calculate matrix to be used in the computation of planar flow
    
    Ixx += ee * (etaNow  * etaNow + r * r/4) ;  
    Iyy += ee * (phiNow * phiNow + r * r/4) ;
    Ixy += ee * etaNow * phiNow; 
    }         

    //Return planar flow.
    return(4 * (Ixx * Iyy -Ixy * Ixy)/((Ixx + Iyy )*(Ixx + Iyy )) );
}


/// Compute angularity jet shape.

double Event::angularity(int a, double R) const {

//Definition 0807.0234v1

  double sum = 0.;
  for (int i = 0; i < size(); ++i){
    double Theta = max (sqrt(particles[i].plain_distance(axis())), TINY);
    double cosTheta = cos(Theta);
    double sinTheta = sqrt(1 - cosTheta * cosTheta);
    sum+= particles[i].pT() * pow(sinTheta, a) * pow(1-cosTheta, 1-a);
  }
  return (1/axis().mCalc() * sum);
}


double Event::thetaS() const {
  return (theta(axis(), particles[0]));
}

double Event::thetaBar () const { 
  double sumTheta = 0.;
  for (int i = 0; i < int(size());++i) 
  sumTheta += sin (theta(axis(), particles[i]));		
  return (sumTheta);	
}


double Event::dipolarity(Vec4 j1, Vec4 j2) const{

  double eta1=j1.eta(), eta2=j2.eta(),phi1=j1.phi(),phi2=j2.phi();
  double aux=0., x2=eta2-eta1, y2=phi2-phi1,dR12=sqrt(j1.plain_distance(j2));
  for (int i = 0; i < int(size());++i){
    double etaNow=particles[i].eta(), phiNow=particles[i].phi(),pTnow=particles[i].pT();
    double x1=etaNow-eta1,y1=phiNow-phi1;
    double Rnow=(1/dR12)*abs(x1*y2-x2*y1);
    aux += pTnow*Rnow*Rnow;
  }
  return (1/(axis().pT()*dR12*dR12)*aux);
}


double Event::dipolarity2(Vec4 j1, Vec4 j2) const{

  double eta1=j1.eta(), eta2=j2.eta(),phi1=j1.phi(),phi2=j2.phi();
  double aux=0., x2=eta2-eta1, y2=phi2-phi1,dR12=sqrt(j1.plain_distance(j2));
  for (int i = 0; i < int(size());++i){
    double etaNow=particles[i].eta(), phiNow=particles[i].phi();
    double x1=etaNow-eta1,y1=phiNow-phi1;
    double Rnow=(1/dR12)*abs(x1*y2-x2*y1);
    aux += Rnow*Rnow;
  }
  return (1/(dR12*dR12)*aux);
}

/// Read Event from a file.

bool Event::eventRead(ifstream& is) {

  string line, tag;
  double pup1, pup2, pup3, pup4, m, pt;
  clear();
  if (!getline(is, line)) return false;
  
  // Read information about jet
  istringstream getpro(line);
  int nupSave, nPileup, nEvt;
  getpro >>nEvt >> nupSave >> nPileup >> pt >> m;
  
  //Read 4-Vec  jet
  if (!getline(is, line)) return false;
  istringstream getall(line);
  getall  >> pup1 >> pup2 >> pup3 >> pup4 >> m;
  
  //Read b-jet 4-Vec
  if (!getline(is, line)) return false;
  istringstream getall2(line);
  getall2  >> pup1 >> pup2 >> pup3 >> pup4 >> m;
  _bJets.push_back( Vec4( pup1, pup2, pup3, pup4));
  
  //Read  another b-jet 4-Vec
  if (!getline(is, line)) return false;
  istringstream getall3(line);
  getall3  >> pup1 >> pup2 >> pup3 >> pup4 >> m;
  _bJets.push_back(Vec4( pup1, pup2, pup3, pup4)); 
  
  for (int ip = 0; ip < nupSave; ++ip) { 
    if (!getline(is, line)) return false;
    istringstream getall(line);
    getall  >> pup1 >> pup2 >> pup3 >> pup4;
    append( Vec4( pup1, pup2, pup3, pup4) );
  }
  return true;
}


bool Event::eventRead(const string& tFile) {
  
  ifstream file(tFile.c_str()); 
  eventRead(file);
  file.close();

  return true;

}

void Event::copyTo(jet_t & jet)  const{

  jet.clear();
  double jetRapidity = axis().eta() ;	  
  double jetPhi =      axis().phi()  ;

  for (int i = 0; i < size(); ++i) {	
    double etaNow = particles[i].eta();
    double phiNow = particles[i].phi();
    double pTnow = particles[i].pT();
    double mNow = particles[i].mCalc(); 
    Vec4 pNow(pTnow * cos(phiNow), pTnow * sin(phiNow), pTnow * sinh(etaNow), 
    sqrt( mNow * mNow + pTnow * pTnow * cosh(etaNow) * cosh(etaNow)) );
    jet.push_back( SingleParticle(pNow, pTnow, particles[i].eta() - jetRapidity, phiNow - jetPhi) );
  }
}

void Event::getFrom(const jet_t & jet) {
     
  clear();
  for (jet_t::const_iterator ip = jet.begin(); ip != jet.end(); ++ip) 
      append( (*ip).p );
     
}
}